"""Filter concepts by concepts, topics, and levels."""

from argparse import ArgumentParser
from typing import NoReturn

from .language.concept import Concept, ConceptId
from .topic.topic import Topic


def filter_concepts(
    concepts: set[Concept],
    topics: set[Topic],
    selected_concepts: list[ConceptId],
    selected_topics: list[str],
    argument_parser: ArgumentParser,
) -> set[Concept] | NoReturn:
    """Filter the concepts by selected concepts, levels, and topics."""
    if selected_concepts:
        concepts = {concept for concept in concepts if concept.concept_id in selected_concepts}
    if selected_topics:
        concept_ids: set[ConceptId] = set()
        for topic in topics:
            if topic.name in selected_topics:
                concept_ids |= topic.concepts
        concepts = {concept for concept in concepts if concept.concept_id in concept_ids}
    if (selected_concepts or selected_topics) and not concepts:
        argument_parser.error("No concepts found that match your selection criteria\n")
    return concepts



"""Labels."""

from __future__ import annotations

from collections.abc import Sequence
from string import punctuation
from typing import Final


class Label(str):
    """Class representing labels for concepts."""

    __slots__ = ()

    # Labels can have one question note and multiple answer notes. The question note is shown before a quiz is
    # presented to the user. The answer notes are shown afterwards. The format is:
    # 'label;question note;answer note 1;answer note 2;...'
    # If the label itself ends with an asterisk it's a colloquial label, i.e. spoken language only.
    NOTE_SEP: Final = ";"
    COLLOQUIAL_POSTFIX: Final = "*"
    QUESTION_NOTE_INDEX: Final = 1
    ANSWER_NOTE_INDEX: Final = 2
    SPELLING_ALTERNATIVES_SEP: Final = "|"

    def __eq__(self, other: object) -> bool:
        """Ignore notes when determining equality."""
        return self.without_notes == Label(other).without_notes

    def __ne__(self, other: object) -> bool:
        """Return whether the labels are not equal."""
        return self.without_notes != Label(other).without_notes

    @property
    def spelling_alternatives(self) -> Labels:
        """Extract the spelling alternatives from the label."""
        return label_factory(self.without_notes.split(self.SPELLING_ALTERNATIVES_SEP))

    @property
    def question_note(self) -> str:
        """Return the label question note."""
        has_question_note = self.count(self.NOTE_SEP) >= self.QUESTION_NOTE_INDEX
        return self.split(self.NOTE_SEP)[self.QUESTION_NOTE_INDEX] if has_question_note else ""

    @property
    def answer_notes(self) -> Sequence[str]:
        """Return the label answer notes."""
        has_answer_notes = self.count(self.NOTE_SEP) >= self.ANSWER_NOTE_INDEX
        return self.split(self.NOTE_SEP)[self.ANSWER_NOTE_INDEX :] if has_answer_notes else ()

    @property
    def without_notes(self) -> str:
        """Return the label without the notes."""
        return self.split(self.NOTE_SEP)[0]

    @property
    def is_colloquial(self) -> bool:
        """Return whether this is a colloquial label."""
        return self.without_notes.endswith(self.COLLOQUIAL_POSTFIX)

    @property
    def is_complete_sentence(self) -> bool:
        """Return whether this is a complete sentence (starts with a capital and has punctuation)."""
        return self.without_notes[0].isupper() and bool(set(self.without_notes) & set(punctuation))

    @property
    def pronounceable(self) -> str:
        """Return the label as text that can be sent to a speech synthesizer."""
        return self.without_notes.rstrip(self.COLLOQUIAL_POSTFIX).replace("'", "")


Labels = tuple[Label, ...]


def label_factory(string: str | list[str]) -> Labels:
    """Instantiate the labels from a string or list of strings."""
    labels = string if isinstance(string, list) else [string]
    return tuple(Label(label) for label in labels if not label.startswith("("))


def meaning_factory(string: str | list[str]) -> Labels:
    """Instantiate the meanings from a string or list of strings."""
    meanings = string if isinstance(string, list) else [string]
    return tuple(
        Label(meaning.removeprefix("(").removesuffix(")"))
        for meaning in meanings
        if not meaning.endswith(Label.COLLOQUIAL_POSTFIX)
    )



"""Topic model."""

from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

from ..language.concept import ConceptId


@dataclass(frozen=True)
class Topic:
    """Topic model."""

    name: str
    _concepts: frozenset[ConceptId] = frozenset()
    topics: frozenset[str] = frozenset()

    instances: ClassVar[dict[str, Topic]] = {}

    def __post_init__(self) -> None:
        """Add the topic to the topic name -> topic mapping."""
        self.instances[self.name] = self

    @property
    def concepts(self) -> frozenset[ConceptId]:
        """Return this topic's concepts."""
        concepts = set(self._concepts)
        for topic in self.topics:
            concepts |= self.instances[topic].concepts
        return frozenset(concepts)



"""Progress model class."""

from collections import deque

from toisto.model.language import Language
from toisto.model.language.concept import Concept

from .quiz import Quiz, Quizzes
from .retention import Retention

ProgressDict = dict[str, dict[str, str | int]]


class Progress:
    """Keep track of progress on quizzes."""

    def __init__(self, progress_dict: ProgressDict, target_language: Language, skip_concepts: int = 5) -> None:
        self.__progress_dict = {key: Retention.from_dict(value) for key, value in progress_dict.items()}
        self.target_language = target_language
        self.__recent_concepts: deque[Concept] = deque(maxlen=skip_concepts)

    def mark_correct_answer(self, quiz: Quiz) -> None:
        """Increase the retention of the quiz."""
        self.__progress_dict.setdefault(quiz.key, Retention()).increase()

    def mark_incorrect_answer(self, quiz: Quiz) -> None:
        """Reset the retention of the quiz."""
        self.__progress_dict.setdefault(quiz.key, Retention()).reset()

    def next_quiz(self, quizzes: Quizzes) -> Quiz | None:
        """Return the next quiz."""
        eligible_quizzes = Quizzes(quiz for quiz in quizzes if self.__is_eligible(quiz))
        quizzes_for_concepts_in_progress = Quizzes(
            quiz for quiz in eligible_quizzes if self.__has_concept_in_progress(quiz, quizzes)
        )
        quizzes_in_progress = Quizzes(quiz for quiz in quizzes_for_concepts_in_progress if self.__in_progress(quiz))
        for potential_quizzes in [quizzes_in_progress, quizzes_for_concepts_in_progress, eligible_quizzes]:
            if unblocked_quizzes := self.__unblocked_quizzes(potential_quizzes, eligible_quizzes, quizzes):
                quiz = unblocked_quizzes.pop()
                self.__recent_concepts.append(quiz.concept.base_concept)
                return quiz
        return None

    def get_retention(self, quiz: Quiz) -> Retention:
        """Return the quiz retention."""
        return self.__progress_dict.get(quiz.key, Retention())

    def __is_eligible(self, quiz: Quiz) -> bool:
        """Return whether the quiz is not silenced and not the current quiz."""
        return quiz.concept.base_concept not in self.__recent_concepts and not self.get_retention(quiz).is_silenced()

    def __has_concept_in_progress(self, quiz: Quiz, quizzes: Quizzes) -> bool:
        """Return whether the quiz's concept has been presented to the user before."""
        quizzes_for_same_concept = quizzes.by_concept(quiz.concept)
        return any(self.__in_progress(quiz_for_same_concept) for quiz_for_same_concept in quizzes_for_same_concept)

    def __in_progress(self, quiz: Quiz) -> bool:
        """Return whether the quiz has been presented to the user before."""
        return quiz.key in self.__progress_dict

    def __unblocked_quizzes(self, potential_quizzes: Quizzes, eligible_quizzes: Quizzes, quizzes: Quizzes) -> Quizzes:
        """Return the quizzes that are not blocked by other quizzes.

        Quiz A is blocked by quiz B if the concept of quiz A is a compound with a root that is quizzed by quiz B.
        """
        return Quizzes(
            quiz
            for quiz in potential_quizzes
            if not self.__root_concepts_have_quizzes(quiz, eligible_quizzes, quizzes)
            and not quiz.is_blocked_by(eligible_quizzes)
        )

    def __root_concepts_have_quizzes(self, quiz: Quiz, eligible_quizzes: Quizzes, quizzes: Quizzes) -> bool:
        """Return whether the quiz's concept has root concepts that have quizzes."""
        target_language = quiz.answer_language if "write" in quiz.quiz_types else quiz.question_language
        return any(
            other_quiz
            for root in quiz.concept.related_concepts.roots(target_language)
            for other_quiz in quizzes.by_concept(root)
            if other_quiz != quiz and other_quiz in eligible_quizzes
        )

    def as_dict(self) -> dict[str, dict[str, int | str]]:
        """Return the progress as dict."""
        return {key: value.as_dict() for key, value in self.__progress_dict.items()}



"""Command to show concepts."""

from itertools import chain

from rich.table import Table

from toisto.model.language import Language
from toisto.model.language.concept import Concept
from toisto.model.language.iana_language_subtag_registry import ALL_LANGUAGES
from toisto.model.language.label import Labels
from toisto.model.topic.topic import Topic
from toisto.ui.text import console


def enumerate_labels(labels: Labels) -> str:
    """Enumerate the labels."""
    return "\n".join(chain.from_iterable(label.spelling_alternatives for label in labels))


def topic_table(
    target_language: Language,
    source_language: Language,
    topic: Topic,
    topics: set[Topic],
    concepts: set[Concept],
) -> Table:
    """Show the concepts of the topic."""
    table = Table(title=f"Topic {topic.name}")
    target_language_name, source_language_name = ALL_LANGUAGES[target_language], ALL_LANGUAGES[source_language]
    for column in (
        target_language_name,
        source_language_name,
        "Grammatical categories",
        "Other topics",
    ):
        table.add_column(column)
    for concept in concepts:
        other_topics = ", ".join(
            sorted(t.name for t in topics if t.name != topic.name and concept.concept_id in t.concepts),
        )
        for leaf_concept in concept.leaf_concepts(target_language):
            target_labels = leaf_concept.labels(target_language)
            source_labels = leaf_concept.labels(source_language)
            table.add_row(
                enumerate_labels(target_labels),
                enumerate_labels(source_labels),
                "/".join(leaf_concept.grammatical_categories()),
                other_topics,
            )
        table.add_section()
    return table


def show_topics(
    target_language: Language,
    source_language: Language,
    topics: set[Topic],
    concepts: set[Concept],
) -> None:
    """Show the concepts by topic."""
    with console.pager():
        for topic in sorted(topics, key=lambda topic: topic.name):
            topic_concepts = {concept for concept in concepts if concept.concept_id in topic.concepts}
            if topic_concepts:
                console.print(topic_table(target_language, source_language, topic, topics, topic_concepts))



"""Quiz factory."""

from dataclasses import dataclass
from itertools import permutations, zip_longest

from ..language import Language
from ..language.concept import Concept, Concepts
from .quiz import GRAMMATICAL_QUIZ_TYPES, Quiz, QuizType, Quizzes


@dataclass(frozen=True)
class QuizFactory:
    """Create quizzes for concepts."""

    target_language: Language
    source_language: Language

    def create_quizzes(self, *concepts: Concept) -> Quizzes:
        """Create quizzes for the concepts."""
        return Quizzes(Quizzes().union(*(self.concept_quizzes(concept, Quizzes()) for concept in concepts)))

    def concept_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create the quizzes for a concept."""
        if concept.answer_only:
            return Quizzes()
        return Quizzes(
            self.composite_concept_quizzes(concept, previous_quizzes)
            | self.leaf_concept_quizzes(concept, previous_quizzes),
        )

    def composite_concept_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create the quizzes for a composite concept."""
        quizzes = Quizzes()
        for constituent_concept in concept.constituents:
            quizzes |= self.concept_quizzes(constituent_concept, Quizzes(quizzes | previous_quizzes))
        quizzes |= self.grammatical_quizzes(concept, Quizzes(quizzes | previous_quizzes))
        return quizzes

    def leaf_concept_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create the quizzes for a leaf concept."""
        translation_quizzes = self.translation_quizzes(concept, previous_quizzes)
        semantic_quizzes = self.semantic_quizzes(concept, Quizzes(translation_quizzes | previous_quizzes))
        return Quizzes(translation_quizzes | semantic_quizzes)

    def translation_quizzes(self, concept: Concept, previous_quizzes: Quizzes | None = None) -> Quizzes:
        """Create the translation quizzes for a concept."""
        previous_quizzes = previous_quizzes or Quizzes()
        read_quizzes = self.read_quizzes(concept, previous_quizzes)
        dictate_quizzes = self.dictate_quizzes(concept, Quizzes(read_quizzes | previous_quizzes))
        write_quizzes = self.write_quizzes(concept, Quizzes(read_quizzes | dictate_quizzes | previous_quizzes))
        interpret_quizzes = self.interpret_quizzes(
            concept,
            Quizzes(read_quizzes | write_quizzes | dictate_quizzes | previous_quizzes),
        )
        return Quizzes(read_quizzes | write_quizzes | dictate_quizzes | interpret_quizzes)

    def read_quizzes(self, concept: Concept, previous_quizzes: Quizzes | None = None) -> Quizzes:
        """Create read quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        if concept.is_composite(target_language):
            return Quizzes()
        target_labels = concept.non_colloquial_labels(target_language)
        source_labels = concept.non_colloquial_labels(source_language)
        if not target_labels or not source_labels:
            return Quizzes()
        blocked_by = tuple(previous_quizzes) if previous_quizzes else ()
        return Quizzes(
            Quiz(concept, target_language, source_language, target_label, source_labels, ("read",), blocked_by)
            for target_label in target_labels
        )

    def write_quizzes(self, concept: Concept, previous_quizzes: Quizzes | None = None) -> Quizzes:
        """Create write quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        if concept.is_composite(target_language):
            return Quizzes()
        target_labels = concept.non_colloquial_labels(target_language)
        source_labels = concept.non_colloquial_labels(source_language)
        if not target_labels or not source_labels:
            return Quizzes()
        blocked_by = tuple(previous_quizzes) if previous_quizzes else ()
        return Quizzes(
            Quiz(concept, source_language, target_language, source_label, target_labels, ("write",), blocked_by)
            for source_label in source_labels
        )

    def dictate_quizzes(self, concept: Concept, previous_quizzes: Quizzes | None = None) -> Quizzes:
        """Create dictation quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        target_labels = concept.non_colloquial_labels(target_language)
        blocked_by = tuple(previous_quizzes) if previous_quizzes else ()
        meanings = concept.meanings(source_language)
        non_colloquial_quizzes = Quizzes(
            Quiz(concept, target_language, target_language, label, (label,), ("dictate",), blocked_by, meanings)
            for label in target_labels
        )
        colloquial_quizzes = Quizzes(
            Quiz(concept, target_language, target_language, label, target_labels, ("dictate",), blocked_by, meanings)
            for label in concept.colloquial_labels(target_language)
        )
        return Quizzes(non_colloquial_quizzes | colloquial_quizzes)

    def interpret_quizzes(self, concept: Concept, previous_quizzes: Quizzes | None = None) -> Quizzes:
        """Create interpret (listen and translate) quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        source_labels = concept.non_colloquial_labels(source_language)
        if not source_labels or concept.is_composite(target_language):
            return Quizzes()
        blocked_by = tuple(previous_quizzes) if previous_quizzes else ()
        meanings = concept.meanings(target_language)
        return Quizzes(
            Quiz(concept, target_language, source_language, label, source_labels, ("interpret",), blocked_by, meanings)
            for label in concept.labels(target_language)
        )

    def grammatical_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create grammatical quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        blocked_by = tuple(previous_quizzes)
        quizzes = Quizzes()
        for question_concept, answer_concept in permutations(concept.leaf_concepts(target_language), r=2):
            quiz_types = grammatical_quiz_types(question_concept, answer_concept)
            if not quiz_types:
                continue
            question_labels = question_concept.non_colloquial_labels(target_language)
            answer_labels = answer_concept.non_colloquial_labels(target_language)
            question_meanings = question_concept.meanings(source_language)
            answer_meanings = answer_concept.meanings(source_language)
            quizzes |= Quizzes(
                Quiz(
                    concept,
                    target_language,
                    target_language,
                    question_label,
                    (answer_label,),
                    quiz_types,
                    blocked_by,
                    question_meanings,
                    answer_meanings,
                )
                for question_label, answer_label in zip(question_labels, answer_labels, strict=False)
                if question_label != answer_label
            )
        return quizzes

    def semantic_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create semantic quizzes for the concept."""
        answer_quizzes = self.related_concept_quizzes(concept, previous_quizzes, "answer", concept.answers)
        antonym_quizzes = self.related_concept_quizzes(
            concept,
            Quizzes(answer_quizzes | previous_quizzes),
            "antonym",
            concept.antonyms,
        )
        return Quizzes(answer_quizzes | antonym_quizzes)

    def related_concept_quizzes(
        self,
        concept: Concept,
        previous_quizzes: Quizzes,
        quiz_type: QuizType,
        related_concepts: Concepts,
    ) -> Quizzes:
        """Create quizzes for the related concepts."""
        if not related_concepts:
            return Quizzes()
        target_language, source_language = self.target_language, self.source_language
        meanings = list(concept.meanings(source_language))
        related_concept_labels = []
        for related_concept in related_concepts:
            meanings.extend(related_concept.meanings(source_language))
            related_concept_labels.extend(list(related_concept.labels(target_language)))
            previous_quizzes |= self.translation_quizzes(related_concept)
        return Quizzes(
            Quiz(
                concept,
                target_language,
                target_language,
                label,
                tuple(related_concept_labels),
                (quiz_type,),
                tuple(previous_quizzes),
                (),
                tuple(meanings),
            )
            for label in concept.non_colloquial_labels(target_language)
        )


def grammatical_quiz_types(concept1: Concept, concept2: Concept) -> tuple[QuizType, ...]:
    """Return the quiz types to change the grammatical category of concept1 into that of concept2.

    For example, to change "I am" into "they are" would mean changing the grammatical number from singular to plural
    and changing the grammatical person from first person to third person. To prevent the quiz from becoming too
    complex ("Give the affirmative past tense plural third person...") we limit the number of quiz types.
    """
    quiz_types = []
    for category1, category2 in zip_longest(concept1.grammatical_categories(), concept2.grammatical_categories()):
        if category1 != category2 and category2 in GRAMMATICAL_QUIZ_TYPES:
            quiz_types.append(GRAMMATICAL_QUIZ_TYPES[category2])
    if set(quiz_types) <= {"feminize", "masculinize", "neuterize", "give third person"}:
        return tuple(quiz_types)
    return tuple(quiz_types) if len(quiz_types) == 1 else tuple()


def create_quizzes(target_language: Language, source_language: Language, *concepts: Concept) -> Quizzes:
    """Create quizzes for the concepts, using the target and source language."""
    return QuizFactory(target_language, source_language).create_quizzes(*concepts)



"""Load concepts from concept files and generate quizzes."""

from argparse import ArgumentParser
from pathlib import Path
from typing import NoReturn

from ..metadata import NAME
from ..model.language.concept import Concept, ConceptId, ConceptIds
from ..model.language.concept_factory import create_concept
from .json_file import load_json


class ConceptIdRegistry:
    """Registry to check the uniqueness of concept identifiers across concept files."""

    def __init__(self, argument_parser: ArgumentParser) -> None:
        self.argument_parser = argument_parser
        self.concept_files_by_concept_id: dict[ConceptId, Path] = {}

    def check_concept_ids(self, concept_ids: ConceptIds, concept_file: Path) -> None:
        """Check that the concept ids are unique."""
        for concept_id in concept_ids:
            self._check_concept_id(concept_id, concept_file)

    def _check_concept_id(self, concept_id: ConceptId, concept_file: Path) -> None:
        """Check that the concept id is unique."""
        if concept_id in self.concept_files_by_concept_id:
            other_concept_file = self.concept_files_by_concept_id[concept_id]
            self.argument_parser.error(
                f"{NAME} cannot read concept file {concept_file}: concept identifier '{concept_id}' also occurs in "
                f"concept file {other_concept_file}.\nConcept identifiers must be unique across concept files.\n",
            )

    def register_concept_ids(self, concept_ids: ConceptIds, concept_file: Path) -> None:
        """Register the concept ids."""
        for concept_id in concept_ids:
            self.concept_files_by_concept_id[concept_id] = concept_file


def load_concepts(
    concept_files: list[Path],
    concept_id_registry: ConceptIdRegistry,
    argument_parser: ArgumentParser,
) -> set[Concept] | NoReturn:
    """Load the concepts from the specified concept files and with the specified levels."""
    all_concepts = set()
    for concept_file in concept_files:
        concepts = []
        try:
            for concept_key, concept_value in load_json(concept_file).items():
                concept = create_concept(concept_key, concept_value)
                concepts.append(concept)
        except Exception as reason:  # noqa: BLE001
            argument_parser.error(f"{NAME} cannot read concept file {concept_file}: {reason}.\n")
        concept_ids = tuple(concept.concept_id for concept in concepts)
        concept_id_registry.check_concept_ids(concept_ids, concept_file)
        concept_id_registry.register_concept_ids(concept_ids, concept_file)
        all_concepts |= set(concepts)
    return all_concepts



"""Store and load progress data."""

from argparse import ArgumentParser
from typing import NoReturn

from ..metadata import NAME, get_progress_filepath
from ..model.language import Language
from ..model.quiz.progress import Progress
from .json_file import dump_json, load_json


def load_progress(target_language: Language, argument_parser: ArgumentParser) -> Progress | NoReturn:
    """Load the progress from the user's home folder."""
    progress_filepath = get_progress_filepath(target_language)
    try:
        progress_dict = load_json(progress_filepath, default={})
    except Exception as reason:  # noqa: BLE001
        return argument_parser.error(
            f"""{NAME} cannot parse the progress information in {progress_filepath}: {reason}.
To fix this, remove or rename {progress_filepath} and start {NAME} again. Unfortunately, this will reset your progress.
Please consider opening a bug report at https://github.com/fniessink/{NAME.lower()}. Be sure to attach the invalid
progress file to the issue.
""",
        )
    return Progress(progress_dict, target_language)


def save_progress(progress: Progress) -> None:
    """Save the progress to the user's home folder."""
    progress_filepath = get_progress_filepath(progress.target_language)
    dump_json(progress_filepath, progress.as_dict())



"""Text to speech."""

import sys
import tempfile
from configparser import ConfigParser
from subprocess import DEVNULL, Popen
from typing import Final

import gtts
from playsound import playsound

MAC_OS_SAY_VOICES: Final = dict(en="Daniel", fi="Satu", nl="Xander")


def _run_command(command: str, *args: str) -> None:
    """Run the command in the background and send output to /dev/null."""
    # Suppress the ruff message: "S603 `subprocess` call: check for execution of untrusted input". Popen should be
    # safe because it's invoked with either "say" or a mp3 play command provided by the user in the config file.
    Popen([command, *args], stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL)  # noqa: S603


def _say_with_macos_say(language: str, text: str, slow: bool) -> None:
    """Say the text with the MacOS say command."""
    voice = MAC_OS_SAY_VOICES[language]
    args = [f"--voice={voice}"]
    if slow:
        args.append("--rate=150")
    args.append(text.replace("'", r"\'"))
    _run_command("say", *args)


def _say_with_google_translate(language: str, text: str, config: ConfigParser, slow: bool) -> None:
    """Say the text with Google Translate say command."""
    try:
        service = gtts.gTTS(text=text, lang=language, slow=slow)
        with tempfile.NamedTemporaryFile(delete=False, suffix=".mp3") as mp3_file:
            service.save(mp3_file.name)
    except gtts.tts.gTTSError as reason:
        message = f"Can't use Google Translate: {reason}"
        raise RuntimeError(message) from reason
    mp3_play_command = config.get("commands", "mp3player")
    if mp3_play_command == "playsound":
        playsound(mp3_file.name)
    else:
        _run_command(mp3_play_command, mp3_file.name)


def say(language: str, text: str, config: ConfigParser, slow: bool = False) -> None:
    """Say the text in the specified language."""
    try:
        _say_with_google_translate(language, text, config, slow)
    except RuntimeError:
        if sys.platform == "darwin":
            _say_with_macos_say(language, text, slow)
        else:
            raise



"""Text matching."""

import string

# Translation table to remove punctuation from strings, except apostrophes and hyphens.
WITHOUT_PUNCTUATION = str.maketrans("", "", string.punctuation)
for char in ["'", "-"]:
    del WITHOUT_PUNCTUATION[ord(char)]


def match(text1: str, *texts: str) -> bool:
    """Return whether the text matches any of the texts."""
    return any(
        text1.strip().translate(WITHOUT_PUNCTUATION) == text2.strip().translate(WITHOUT_PUNCTUATION) for text2 in texts
    )



"""Concept classes."""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import ClassVar, NewType, cast, get_args

from . import Language
from .grammar import GrammaticalCategory
from .label import Labels

ConceptId = NewType("ConceptId", str)
ConceptIds = tuple[ConceptId, ...]


@dataclass(frozen=True)
class RelatedConcepts:
    """Class representing the relations of a concept with other concepts.

    Concepts can be composites, meaning they consists of subconcepts (called constituent concepts). For example, a noun
    concept may have a singular and plural form. These forms are represented as constituent concepts of the parent
    concept. The parent and constituent attributes keep track of these relations.

    Another relation type that may exist between concepts is roots. Compound concepts such as 'blackboard' contain two
    roots, 'black' and 'board'. The concept for the compound concept refers to its roots with the roots attribute. The
    roots relation can also be used for sentences, in which case the individual words of a sentence are the roots.

    The antonym relation is used to capture opposites. For example 'bad' has 'good' as antonym and 'good' has 'bad'
    as antonym.

    The answer relation is used to specify possible answers to questions. Using the answer relation it is possible to
    specify that, for example, 'Do you like ice cream?' has the yes and no concepts as possible answers.

    NOTE: This class keeps track of the related concepts using their concept identifier (ConceptId) and only when
    the client asks for a concept is the concept instance looked up in the concept registry (Concept.instances). This
    prevents the need for a second pass after instantiating concepts from the concept files to create the relations.
    """

    _parent: ConceptId | None
    _constituents: ConceptIds
    _roots: dict[Language, ConceptIds] | ConceptIds  # Tuple if all languages have the same roots
    _antonyms: ConceptIds
    _answers: ConceptIds

    @property
    def parent(self) -> Concept | None:
        """Return the parent concept."""
        return self._get_concepts(self._parent)[0] if self._parent else None

    @property
    def constituents(self) -> Concepts:
        """Return the constituent concepts."""
        return self._get_concepts(*self._constituents)

    @property
    def antonyms(self) -> Concepts:
        """Return the antonym (opposite) concepts."""
        return self._get_concepts(*self._antonyms)

    @property
    def answers(self) -> Concepts:
        """Return the answers to the question that this concept poses."""
        return self._get_concepts(*self._answers)

    def roots(self, language: Language) -> Concepts:
        """Return the root concepts, for the specified language."""
        concept_ids_of_roots = self._roots.get(language, ()) if isinstance(self._roots, dict) else self._roots
        return self._get_concepts(*concept_ids_of_roots)

    def _get_concepts(self, *concept_ids: ConceptId) -> Concepts:
        """Return the concepts with the given concept ids."""
        return tuple(Concept.instances[concept_id] for concept_id in concept_ids if concept_id in Concept.instances)


@dataclass(frozen=True)
class Concept:
    """Class representing language concepts.

    A concept is either a composite or a leaf concept. Composite concepts have have two or more constituent concepts,
    representing different grammatical categories, for example singular and plural forms. Leaf concepts have labels
    in different languages.
    """

    concept_id: ConceptId
    _labels: dict[Language, Labels]
    _meanings: dict[Language, Labels]
    related_concepts: RelatedConcepts
    answer_only: bool

    instances: ClassVar[dict[ConceptId, Concept]] = {}

    def __post_init__(self) -> None:
        """Add the concept to the concept id -> concept mapping."""
        self.instances[self.concept_id] = self

    def __hash__(self) -> int:
        """Return the concept hash."""
        return hash(self.concept_id)

    def __getattr__(self, attribute: str) -> Concepts:
        """Forward properties to related concepts."""
        return getattr(self.related_concepts, attribute)

    @cached_property
    def base_concept(self) -> Concept:
        """Return the base concept of this concept."""
        return self.related_concepts.parent.base_concept if self.related_concepts.parent else self

    def leaf_concepts(self, language: Language) -> Generator[Concept, None, None]:
        """Return this concept's leaf concepts, or self if this concept is a leaf concept."""
        if self.is_composite(language):
            for concept in self.constituents:
                yield from concept.leaf_concepts(language)
        else:
            yield self

    def labels(self, language: Language) -> Labels:
        """Return the labels for the language."""
        if self.is_composite(language):
            return tuple(chain.from_iterable([concept.labels(language) for concept in self.constituents]))
        return self._labels.get(language, Labels())

    def colloquial_labels(self, language: Language) -> Labels:
        """Return the colloquial labels for the language."""
        return Labels(label for label in self.labels(language) if label.is_colloquial)

    def non_colloquial_labels(self, language: Language) -> Labels:
        """Return the non-colloquial labels for the language."""
        return Labels(label for label in self.labels(language) if not label.is_colloquial)

    def meanings(self, language: Language) -> Labels:
        """Return the meanings of the concept in the specified language."""
        if self.is_composite(language):
            return tuple(chain.from_iterable([concept.meanings(language) for concept in self.constituents]))
        return self._meanings.get(language, Labels())

    def grammatical_categories(self) -> tuple[GrammaticalCategory, ...]:
        """Return the grammatical categories of this concept."""
        keys = self.concept_id.split("/")
        return tuple(cast(GrammaticalCategory, key) for key in keys if key in get_args(GrammaticalCategory))

    def is_composite(self, language: Language) -> bool:
        """Return whether this concept is composite."""
        return self._labels.get(language) is None


Concepts = tuple[Concept, ...]



"""Config file parser."""

import sys
from argparse import ArgumentParser
from collections.abc import Iterable
from configparser import ConfigParser, Error
from enum import Enum
from pathlib import Path
from typing import Final, NoReturn

from toisto.model.language.iana_language_subtag_registry import ALL_LANGUAGES

# The schema for the config file. Top-level keys are sections, values are a dict per option with the key being the
# option name and the value being a tuple of a specifier and the allowed option values.


class Quantifier(Enum):
    """Quantifier enumeration."""

    ANY = "any"
    ONE_OF = "one of"


CONFIG_SCHEMA: Final[dict[str, dict[str, tuple[Quantifier, Iterable]]]] = dict(
    languages=dict(
        target=(Quantifier.ONE_OF, ALL_LANGUAGES.keys()),
        source=(Quantifier.ONE_OF, ALL_LANGUAGES.keys()),
    ),
    commands=dict(mp3player=(Quantifier.ANY, [])),
)
CONFIG_FILENAME = Path("~/.toisto.cfg").expanduser()


class ConfigSchemaValidator:
    """Class to validate a config against the schema."""

    def __init__(self, config_parser: ConfigParser, argument_parser: ArgumentParser, config_filename: Path) -> None:
        self._argument_parser = argument_parser
        self._config_parser = config_parser
        self._config_filename = config_filename

    def _error(self, message: str) -> NoReturn:
        """Report the error and exit."""
        self._argument_parser.error(f"While reading from '{self._config_filename}': {message}")

    def validate(self) -> None | NoReturn:  # type: ignore[return]
        """Validate the config file against the schema."""
        for section in self._config_parser.sections():
            self._validate_section(section)

    def _validate_section(self, section: str) -> None | NoReturn:  # type: ignore[return]
        """Validate the section, including its options."""
        if section not in (allowed_sections := CONFIG_SCHEMA.keys()):
            self._error(f"unknown section '{section}'. Allowed sections are: {', '.join(allowed_sections)}.")
        for option in self._config_parser[section]:
            self._validate_option(section, option)

    def _validate_option(self, section: str, option: str) -> None | NoReturn:  # type: ignore[return]
        """Validate the option, including its value(s)."""
        if option not in (allowed_options := CONFIG_SCHEMA[section].keys()):
            self._error(
                f"unknown option '{option}' in section '{section}'. Allowed options are: {', '.join(allowed_options)}.",
            )
        specifier, allowed_values = CONFIG_SCHEMA[section][option]
        if specifier == Quantifier.ANY:
            return  # type: ignore[return-value]
        value = self._config_parser.get(section, option)
        if value not in allowed_values:
            self._error(
                f"unknown value '{value}' for option '{option}' in section '{section}'. "
                f"Allowed values are {specifier.value}: {', '.join(allowed_values)}.",
            )


def read_config(argument_parser: ArgumentParser, config_filename: Path = CONFIG_FILENAME) -> ConfigParser | NoReturn:
    """Read the config file, validate it, and exit with an error message if it doesn't pass."""
    parser = ConfigParser()
    try:
        with config_filename.open("r", encoding="utf-8") as config_file:
            parser.read_file(config_file)
    except FileNotFoundError:
        pass
    except (OSError, Error) as reason:
        argument_parser.error(str(reason))
    ConfigSchemaValidator(parser, argument_parser, config_filename).validate()
    _add_defaults(parser)
    return parser


def default_config() -> ConfigParser:
    """Return the default configuration."""
    parser = ConfigParser()
    _add_defaults(parser)
    return parser


def _add_defaults(parser: ConfigParser) -> None:
    """Add the default configuration to the parser."""
    if parser.get("commands", "mp3player", fallback=None) is None:
        if "commands" not in parser.sections():
            parser.add_section("commands")
        defaults = dict(darwin="afplay", linux="mpg123 --quiet")
        parser["commands"]["mp3player"] = defaults.get(sys.platform, "playsound")



"""Utitity functions."""

from collections.abc import Callable, Sequence
from typing import TypeVar

T = TypeVar("T")


def first(sequence: Sequence[T], where: Callable[[T], bool] = lambda _item: True) -> T:
    """Return the first item in the sequence."""
    return next(item for item in sequence if where(item))



"""Main module for the application."""

import logging
from contextlib import suppress
from pathlib import Path

with suppress(ImportError):
    import readline  # noqa: F401 `readline` imported but unused

# Suppress warning messages printed by the playsound module.
logging.getLogger().setLevel(logging.ERROR)

from .command.practice import practice
from .command.show_progress import show_progress
from .command.show_topics import show_topics
from .metadata import CONCEPT_JSON_FILES, TOPIC_FILES, latest_version
from .model.filter import filter_concepts
from .model.quiz.quiz_factory import create_quizzes
from .persistence.concepts import ConceptIdRegistry, load_concepts
from .persistence.config import default_config, read_config
from .persistence.progress import load_progress
from .persistence.topics import load_topics
from .ui.cli import create_argument_parser, parse_arguments
from .ui.text import show_welcome


def main() -> None:
    """Run the main program."""
    argument_parser = create_argument_parser(default_config())
    config = read_config(argument_parser)
    registry = ConceptIdRegistry(argument_parser)
    concepts = load_concepts(CONCEPT_JSON_FILES, registry, argument_parser)
    topics = load_topics(TOPIC_FILES, argument_parser)
    argument_parser = create_argument_parser(config, concepts, topics)
    args = parse_arguments(argument_parser)
    extra_concept_files = [Path(concept_file) for concept_file in args.concept_file]
    concepts |= load_concepts(extra_concept_files, registry, argument_parser)
    extra_topic_files = [Path(topic_file) for topic_file in args.topic_file]
    topics |= load_topics(extra_topic_files, argument_parser)
    concepts = filter_concepts(concepts, topics, args.concept, args.topic, argument_parser)
    quizzes = create_quizzes(args.target_language, args.source_language, *concepts)
    progress = load_progress(args.target_language, argument_parser)
    match args.command:
        case "topics":
            show_topics(args.target_language, args.source_language, topics, concepts)
        case "progress":
            show_progress(args.target_language, quizzes, progress, args.sort)
        case _:  # Default command is "practice"
            show_welcome(latest_version())
            practice(quizzes, progress, config)



"""Load topics concepts from topic files."""

from argparse import ArgumentParser
from pathlib import Path
from typing import NoReturn, TypedDict, cast

from ..metadata import NAME
from ..model.language.concept import ConceptId
from ..model.topic.topic import Topic
from .json_file import load_json


class TopicJSON(TypedDict):
    """Topic JSON."""

    name: str
    concepts: list[ConceptId]
    topics: list[str]


def load_topics(topic_files: list[Path], argument_parser: ArgumentParser) -> set[Topic] | NoReturn:
    """Load the topics from the specified topic files."""
    topics = set()
    try:
        for topic_file in topic_files:
            topic_json = cast(TopicJSON, load_json(topic_file))
            topics.add(
                Topic(topic_json["name"], frozenset(topic_json["concepts"]), frozenset(topic_json.get("topics", []))),
            )
    except Exception as reason:  # noqa: BLE001
        argument_parser.error(f"{NAME} cannot read topic file {topic_file}: {reason}.\n")
    return topics



"""Output for the user."""

import sys
from typing import Final

from rich.console import Console
from rich.panel import Panel
from rich.theme import Theme

from ..metadata import CHANGELOG_URL, NAME, VERSION
from ..model.language.label import Label
from ..model.quiz.quiz import Quiz
from .dictionary import DICTIONARY_URL, linkify_and_enumerate
from .diff import colored_diff

theme: Final = Theme(dict(secondary="grey69", quiz="medium_purple1", inserted="bright_green", deleted="bright_red"))

console = Console(theme=theme)

LINK_KEY: Final = "⌘ (the command key)" if sys.platform == "darwin" else "Ctrl (the control key)"

WELCOME: Final = f"""👋 Welcome to [underline]{NAME} [white not bold]v{VERSION}[/white not bold][/underline]!

Practice as many words and phrases as you like, for as long as you like.

[secondary]{NAME} quizzes you on words and phrases repeatedly. Each time you answer
a quiz correctly, {NAME} will wait longer before repeating it. If you
answer incorrectly, you get one additional attempt to give the correct
answer. If the second attempt is not correct either, {NAME} will reset
the quiz interval.

How does it work?
● To answer a quiz: type the answer, followed by Enter.
● To repeat the spoken text: type Enter without answer.
● To skip to the answer immediately: type ?, followed by Enter.
● To read more about an [link={DICTIONARY_URL}/underlined]underlined[/link] word: keep {LINK_KEY} pressed
  while clicking the word. Not all terminals may support this.
● To quit: type Ctrl-C or Ctrl-D.
[/secondary]"""

NEWS: Final = (
    f"🎉 {NAME} [white not bold]{{0}}[/white not bold] is [link={CHANGELOG_URL}]available[/link]. "
    f"Upgrade with [code]pipx upgrade {NAME}[/code]."
    ""
)

DONE: Final = f"""👍 Good job. You're done for now. Please come back later or try a different topic.
[secondary]Type `{NAME.lower()} -h` for more information.[/secondary]
"""

TRY_AGAIN: Final = "⚠️  Incorrect. Please try again."

TRY_AGAIN_IN_ANSWER_LANGUAGE: Final = "⚠️  Incorrect. Please try again, in [bold]%(language)s[/bold]."

CORRECT: Final = "✅ Correct.\n"

INCORRECT: Final = "❌ Incorrect. "


def feedback_correct(guess: Label, quiz: Quiz) -> str:
    """Return the feedback about a correct result."""
    return CORRECT + meaning(quiz) + other_answers(guess, quiz) + answer_notes(quiz)


def feedback_incorrect(guess: Label, quiz: Quiz) -> str:
    """Return the feedback about an incorrect result."""
    if guess == "?":
        label = "The correct answer is" if len(quiz.answers) == 1 else "The correct answers are"
        feedback = f"{label} {linkify_and_enumerate(*quiz.answers)}.\n" + meaning(quiz)
    else:
        evaluation = "" if guess == "?" else INCORRECT
        label = f'{evaluation}The correct answer is "{colored_diff(guess, quiz.answer)}".\n'
        feedback = label + meaning(quiz) + other_answers(quiz.answer, quiz)
    return feedback + answer_notes(quiz)


def meaning(quiz: Quiz) -> str:
    """Return the quiz's meaning, if any."""
    if quiz.question_meanings and quiz.answer_meanings:
        question_meanings = linkify_and_enumerate(*quiz.question_meanings)
        answer_meanings = linkify_and_enumerate(*quiz.answer_meanings)
        meanings = f"{question_meanings} > {answer_meanings}"
    else:
        meanings = linkify_and_enumerate(*(quiz.question_meanings + quiz.answer_meanings))
    return f"[secondary]Meaning {meanings}.[/secondary]\n" if meanings else ""


def other_answers(guess: Label, quiz: Quiz) -> str:
    """Return the quiz's other answers, if any."""
    if answers := quiz.other_answers(guess):
        label = "Another correct answer is" if len(answers) == 1 else "Other correct answers are"
        return f"""[secondary]{label} {linkify_and_enumerate(*answers)}.[/secondary]\n"""
    return ""


def answer_notes(quiz: Quiz) -> str:
    """Return the answer notes, if any."""
    notes = quiz.answer_notes
    if len(notes) == 0:
        return ""
    if len(notes) == 1:
        return f"[secondary]Note: {notes[0]}.[/secondary]\n"
    return "\n".join(["[secondary]Notes:"] + [f"- {note}." for note in notes] + ["[/secondary]\n"])


def instruction(quiz: Quiz) -> str:
    """Return the instruction for the quiz."""
    return f"[quiz]{quiz.instruction}:[/quiz]"


def show_welcome(latest_version: str | None) -> None:
    """Show the welcome message."""
    console.print(WELCOME)
    if latest_version and latest_version.strip("v") > VERSION:
        console.print(Panel(NEWS.format(latest_version), expand=False))
        console.print()



"""Load and dump JSON."""

import errno
import json
import os
from pathlib import Path


def load_json(json_file_path: Path, default: dict | None = None) -> dict:
    """Load the JSON from the file. Return default if file does not exist."""
    if json_file_path.exists():
        with json_file_path.open(encoding="utf-8") as json_file:
            return json.load(json_file)
    if default is None:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(json_file_path))
    return default or {}


def dump_json(json_file_path: Path, contents: dict | list) -> None:
    """Dump the JSON into the file."""
    with json_file_path.open("w", encoding="utf-8") as json_file:
        json.dump(contents, json_file)
